@@ -8,7 +8,7 @@
 //! smaller code but less useful error messages.
 //!
 //! The `std` feature controls whether [`error::Error`](error::Error) implements
-//! [`std::error::Error`](std::error::Error), which it cannot do in a `no_std` environment.
+//! `std::error::Error`, which it cannot do in a `no_std` environment.
 
 #![cfg_attr(not(feature = "std"), no_std)]
 #![warn(
@@ -86,8 +86,6 @@ impl FromStr for Address {
 
 impl decode::Decode<'_> for Address {
 	fn decode(d: &mut decode::Decoder<'_>) -> Result<Self, decode::Error> {
-		use core::convert::TryInto;
-
 		// Check the datatype of the next item.
 		let mut datatype = d.datatype()?;
 
@@ -95,7 +93,7 @@ impl decode::Decode<'_> for Address {
 		if datatype == Type::Tag {
 			let tag = d.tag()?;
 			if tag != Tag::Unassigned(39) {
-				return Err(decode::Error::Message("expected tag 39"));
+				return Err(decode::Error::message("expected tag 39"));
 			}
 			datatype = d.datatype()?;
 		}
@@ -107,17 +105,17 @@ impl decode::Decode<'_> for Address {
 			Type::Bytes => {
 				let b = d.bytes()?;
 				Ok(Self::from_bytes(b.try_into().map_err(|_| {
-					decode::Error::Message("expected 16 bytes")
+					decode::Error::message("expected 16 bytes")
 				})?))
 			}
 			Type::String => {
 				let s = d.str()?;
 				Ok(
 					Self::from_str(s)
-						.map_err(|_| decode::Error::Message("expected UUID string"))?,
+						.map_err(|_| decode::Error::message("expected UUID string"))?,
 				)
 			}
-			_ => Err(decode::Error::Message("expected byte or UTF-8 string")),
+			_ => Err(decode::Error::message("expected byte or UTF-8 string")),
 		}
 	}
 }
@@ -154,7 +152,7 @@ macro_rules! panic_or_trap {
 #[macro_export]
 macro_rules! panic_or_trap {
 	($message: literal) => {
-		unsafe { core::arch::wasm32::unreachable() }
+		core::arch::wasm32::unreachable()
 	};
 }
 
